@@ -189,11 +189,6 @@ class Predictor:
 
             from_ds = getDS(from_data)
 
-            # Set user-provided subtypes
-            from_ds.set_subtypes(
-                advanced_args.get('subtypes', {})
-            )
-
             transaction_type = TRANSACTION_LEARN
 
             sample_for_analysis, disable_lightwood_transform_cache = _get_memory_optimizations(from_ds.df)
