@@ -1,5 +1,6 @@
-from mindsdb_native.libs.constants.mindsdb import *
 from mindsdb_native.libs.helpers.general_helpers import *
+from mindsdb_native.libs.helpers.confidence_helpers import get_numerical_conf_range, get_categorical_conf
+from mindsdb_native.libs.helpers.conformal_helpers import restore_icp_state, clear_icp_state
 from mindsdb_native.libs.data_types.transaction_data import TransactionData
 from mindsdb_native.libs.data_types.transaction_output_data import (
     PredictTransactionOutputData,
@@ -8,7 +9,6 @@ from mindsdb_native.libs.data_types.transaction_output_data import (
 from mindsdb_native.libs.data_types.mindsdb_logger import log
 from mindsdb_native.config import CONFIG
 
-from lightwood.api.predictor import Predictor
 
 import _thread
 import traceback
@@ -83,23 +83,15 @@ class Transaction:
         try:
             with open(icp_fn, 'rb') as fp:
                 self.hmd['icp'] = dill.load(fp)
-                # restore MDB predictors in ICP objects
                 for col in self.lmd['predict_columns']:
-                    try:
-                        self.hmd['icp'][col].nc_function.model.model = self.session.transaction.model_backend.predictor
-                    except AttributeError:
-                        model_path = os.path.join(CONFIG.MINDSDB_STORAGE_PATH, self.hmd['name'], 'lightwood_data')
-                        self.hmd['icp'][col].nc_function.model.model = Predictor(load_from_path=model_path)
-
-                    # restore model in normalizer
-                    if self.hmd['icp'][col].nc_function.normalizer is not None:
-                        self.hmd['icp'][col].nc_function.normalizer.model = self.hmd['icp'][col].nc_function.model.model
+                    restore_icp_state(col, self.hmd, self.session)
 
         except FileNotFoundError as e:
-            self.hmd['icp'] = {'active': False}
+            self.hmd['icp'] = {'__mdb_active': False}
             self.log.warning(f'Could not find mindsdb conformal predictor.')
 
         except Exception as e:
+            self.hmd['icp'] = {'__mdb_active': False}
             self.log.error(e)
             self.log.error(f'Could not load mindsdb conformal predictor in the file: {icp_fn}')
 
@@ -136,27 +128,27 @@ class Transaction:
             self.log.error(traceback.format_exc())
             self.log.error(f'Could not save mindsdb heavy metadata in the file: {fn}')
 
-        if 'icp' in self.hmd.keys() and self.hmd['icp']['active']:
+        if 'icp' in self.hmd.keys() and self.hmd['icp']['__mdb_active']:
             icp_fn = os.path.join(CONFIG.MINDSDB_STORAGE_PATH, self.hmd['name'], 'icp.pickle')
+            mdb_predictors = {}
             try:
-                mdb_predictors = {}
                 with open(icp_fn, 'wb') as fp:
-                    # clear data cache
                     for key in self.hmd['icp'].keys():
-                        if key != 'active':
-                            mdb_predictors[key] = self.hmd['icp'][key].nc_function.model.model
-                            self.hmd['icp'][key].nc_function.model.model = None
-                            self.hmd['icp'][key].nc_function.model.last_x = None
-                            self.hmd['icp'][key].nc_function.model.last_y = None
-                            if self.hmd['icp'][key].nc_function.normalizer is not None:
-                                self.hmd['icp'][key].nc_function.normalizer.model = None
+                        if key != '__mdb_active':
+                            mdb_predictors[key] = {}
+                            for group, icp in self.hmd['icp'][key].items():
+                                if group not in ['__mdb_groups', '__mdb_group_keys']:
+                                    mdb_predictors[key][group] = icp.nc_function.model.model
+                                    clear_icp_state(icp.nc_function)
 
                     dill.dump(self.hmd['icp'], fp, protocol=dill.HIGHEST_PROTOCOL)
 
                     # restore predictor in ICP
                     for key in self.hmd['icp'].keys():
-                        if key != 'active':
-                            self.hmd['icp'][key].nc_function.model.model = mdb_predictors[key]
+                        if key != '__mdb_active':
+                            for group, icp in self.hmd['icp'][key].items():
+                                if group not in ['__mdb_groups', '__mdb_group_keys']:
+                                    icp.nc_function.model.model = mdb_predictors[key][group]
 
             except Exception as e:
                 self.log.error(e)
@@ -356,94 +348,136 @@ class PredictTransaction(Transaction):
             else:
                 output_data[col] = list(predictions_df[col])
 
-        # confidence estimation
-        if self.hmd['icp']['active'] and not self.lmd['quick_predict']:
-            self.lmd['all_conformal_ranges'] = {}
-            icp_X = deepcopy(predictions_df)
+        # confidence estimation using calibrated inductive conformal predictors (ICPs)
+        if self.hmd['icp']['__mdb_active'] and not self.lmd['quick_predict']:
 
-            if self.lmd['tss']['is_timeseries']:
-                icp_X, _, _, _ = self.model_backend._ts_reshape(icp_X)  # TODO: avoid inefficient reshaping
+            icp_X = deepcopy(self.input_data.cached_pred_df)
 
-            for col in self.lmd['columns_to_ignore'] + self.lmd['predict_columns']:
+            # replace observed data w/predictions
+            for col in self.lmd['predict_columns']:
+                if col in icp_X.columns:
+                    preds = self.hmd['predictions'][col]
+                    if self.lmd['tss']['is_timeseries'] and self.lmd['tss']['nr_predictions'] > 1:
+                        preds = [p[0] for p in preds]
+                    icp_X[col] = preds
+
+            # erase ignorable columns
+            for col in self.lmd['columns_to_ignore']:
                 if col in icp_X.columns:
                     icp_X.pop(col)
 
+            # get confidence bounds for each target
             for predicted_col in self.lmd['predict_columns']:
-                output_data[predicted_col] = list(self.hmd['predictions'][predicted_col])
                 output_data[f'{predicted_col}_confidence'] = [None] * len(output_data[predicted_col])
                 output_data[f'{predicted_col}_confidence_range'] = [[None, None]] * len(output_data[predicted_col])
 
-                if self.hmd['icp'].get(predicted_col, False):
-                    typing_info = self.lmd['stats_v2'][predicted_col]['typing']
+                typing_info = self.lmd['stats_v2'][predicted_col]['typing']
+                is_numerical = typing_info['data_type'] == DATA_TYPES.NUMERIC or \
+                               (typing_info['data_type'] == DATA_TYPES.SEQUENTIAL and
+                                DATA_TYPES.NUMERIC in typing_info['data_type_dist'].keys())
+
+                is_categorical = (typing_info['data_type'] == DATA_TYPES.CATEGORICAL or
+                                 (typing_info['data_type'] == DATA_TYPES.SEQUENTIAL and
+                                  DATA_TYPES.CATEGORICAL in typing_info['data_type_dist'].keys())) and \
+                                  typing_info['data_subtype'] != DATA_SUBTYPES.TAGS
+
+                if (is_numerical or is_categorical) and self.hmd['icp'].get(predicted_col, False):
+
+                    # reorder DF index
+                    index = self.hmd['icp'][predicted_col]['__default'].index.values
+                    index = np.append(index, predicted_col) if predicted_col not in index else index
+                    icp_X = icp_X.reindex(columns=index)  # important, else bounds can be invalid
+
+                    # only one normalizer, even if it's a grouped time series task
+                    normalizer = self.hmd['icp'][predicted_col]['__default'].nc_function.normalizer
+                    if normalizer:
+                        normalizer.prediction_cache = self.hmd['predictions'].get(f'{predicted_col}_selfaware_scores',
+                                                                                  None)
+                        icp_X['__mdb_selfaware_scores'] = normalizer.prediction_cache
+
+                    # get ICP predictions
+                    result = pd.DataFrame(index=icp_X.index, columns=['lower', 'upper', 'significance'])
+
+                    # base ICP
                     X = deepcopy(icp_X)
 
-                    # preserve order that the ICP expects, else bounds are useless
-                    X = X.reindex(columns=self.hmd['icp'][predicted_col].index.values)
+                    # get all possible ranges
+                    if self.lmd['tss']['is_timeseries'] and self.lmd['tss']['nr_predictions'] > 1 and is_numerical:
 
-                    normalizer = self.hmd['icp'][predicted_col].nc_function.normalizer
-                    if normalizer:
-                        normalizer.prediction_cache = self.hmd['predictions']
+                        # bounds in time series are only given for the first forecast
+                        self.hmd['icp'][predicted_col]['__default'].nc_function.model.prediction_cache = \
+                            [p[0] for p in output_data[predicted_col]]
+                        all_confs = self.hmd['icp'][predicted_col]['__default'].predict(X.values)
 
-                    # numerical
-                    if typing_info['data_type'] == DATA_TYPES.NUMERIC or \
-                            (typing_info['data_type'] == DATA_TYPES.SEQUENTIAL and
-                                DATA_TYPES.NUMERIC in typing_info['data_type_dist'].keys()):
-                        std_tol = 1
-                        tolerance = self.lmd['stats_v2'][predicted_col]['train_std_dev'] * std_tol
-                        if self.lmd['tss']['is_timeseries'] and self.lmd['tss']['nr_predictions'] > 1:
-                            # bounds in time series are only given for the first forecast
-                            self.hmd['icp'][predicted_col].nc_function.model.prediction_cache = \
-                                [p[0] for p in output_data[predicted_col]]
-                        else:
-                            self.hmd['icp'][predicted_col].nc_function.model.prediction_cache = output_data[predicted_col]
-                        self.lmd['all_conformal_ranges'][predicted_col] = self.hmd['icp'][predicted_col].predict(X.values)
-
-                        for sample_idx in range(self.lmd['all_conformal_ranges'][predicted_col].shape[0]):
-                            sample = self.lmd['all_conformal_ranges'][predicted_col][sample_idx, :, :]
-                            for idx in range(sample.shape[1]):
-                                significance = (99 - idx) / 100
-                                diff = sample[1, idx] - sample[0, idx]
-                                if diff <= tolerance:
-                                    output_data[f'{predicted_col}_confidence'][sample_idx] = significance
-                                    conf_range = list(sample[:, idx])
-
-                                    if self.lmd['stats_v2'][predicted_col].get('positive_domain', False):
-                                        conf_range[0] = max(0.0, conf_range[0])
-
-                                    output_data[f'{predicted_col}_confidence_range'][sample_idx] = conf_range
-                                    break
-                            else:
-                                output_data[f'{predicted_col}_confidence'][sample_idx] = 0.9901  # default
-                                bounds = sample[:, 0]
-                                sigma = (bounds[1] - bounds[0]) / 2
-                                conf_range = [bounds[0] - sigma, bounds[1] + sigma]
-
-                                if self.lmd['stats_v2'][predicted_col].get('positive_domain', False):
-                                    conf_range[0] = max(0.0, conf_range[0])
-
-                                output_data[f'{predicted_col}_confidence_range'][sample_idx] = conf_range
+                    elif is_numerical:
+                        self.hmd['icp'][predicted_col]['__default'].nc_function.model.prediction_cache = \
+                            output_data[predicted_col]
+                        all_confs = self.hmd['icp'][predicted_col]['__default'].predict(X.values)
 
                     # categorical
-                    elif typing_info['data_type'] == DATA_TYPES.CATEGORICAL or \
-                            (typing_info['data_type'] == DATA_TYPES.SEQUENTIAL and
-                                DATA_TYPES.CATEGORICAL in typing_info['data_type_dist'].keys()):
-                        if self.lmd['stats_v2'][predicted_col]['typing']['data_subtype'] != DATA_SUBTYPES.TAGS:
-                            significances = list(range(20)) + list(range(20, 100, 10))  # max permitted error rate
-                            self.hmd['icp'][predicted_col].nc_function.model.prediction_cache = output_data[f'{predicted_col}_class_distribution']
-                            all_ranges = np.array(
-                                [self.hmd['icp'][predicted_col].predict(X.values, significance=s / 100)
-                                    for s in significances])
-                            self.lmd['all_conformal_ranges'][predicted_col] = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)
+                    else:
+                        self.hmd['icp'][predicted_col]['__default'].nc_function.model.prediction_cache = \
+                            output_data[f'{predicted_col}_class_distribution']
+
+                        conf_candidates = list(range(20)) + list(range(20, 100, 10))
+                        all_ranges = np.array(
+                            [self.hmd['icp'][predicted_col]['__default'].predict(X.values, significance=s / 100)
+                             for s in conf_candidates])
+                        all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)
+
+                    # convert (B, 2, 99) into (B, 2) given width or error rate constraints
+                    if is_numerical:
+                        significances, confs = get_numerical_conf_range(all_confs, predicted_col,
+                                                                        self.lmd['stats_v2'])
+                        result.loc[X.index, 'lower'] = confs[:, 0]
+                        result.loc[X.index, 'upper'] = confs[:, 1]
+                    else:
+                        conf_candidates = list(range(20)) + list(range(20, 100, 10))
+                        significances = get_categorical_conf(all_confs, conf_candidates)
+
+                    result.loc[X.index, 'significance'] = significances
+
+                    # grouped time series, we replace bounds in rows that have a trained ICP
+                    if self.hmd['icp'][predicted_col].get('__mdb_groups', False):
+                        icps = self.hmd['icp'][predicted_col]
+                        group_keys = icps['__mdb_group_keys']
+
+                        for group in icps['__mdb_groups']:
+
+                            # filter rows by group
+                            X = deepcopy(icp_X)
+                            for key, val in zip(group_keys, group):
+                                X = X[X[key] == val]
+
+                            if X.size > 0:
+                                # set ICP caches
+                                icps[frozenset(group)].nc_function.model.prediction_cache = X.pop(predicted_col).values
+                                if icps[frozenset(group)].nc_function.normalizer:
+                                    icps[frozenset(group)].nc_function.normalizer.prediction_cache = X.pop('__mdb_selfaware_scores').values
+
+                                # predict and get confidence level given width or error rate constraints
+                                if is_numerical:
+                                    all_confs = icps[frozenset(group)].predict(X.values)
+                                    significances, confs = get_numerical_conf_range(all_confs, predicted_col,
+                                                                                    self.lmd['stats_v2'],
+                                                                                    group=frozenset(group))
+                                    result.loc[X.index, 'lower'] = confs[:, 0]
+                                    result.loc[X.index, 'upper'] = confs[:, 1]
 
-                            for sample_idx in range(self.lmd['all_conformal_ranges'][predicted_col].shape[0]):
-                                sample = self.lmd['all_conformal_ranges'][predicted_col][sample_idx, :, :]
-                                for idx in range(sample.shape[1]):
-                                    conf = (99 - significances[idx]) / 100
-                                    if np.sum(sample[:, idx]) == 1:
-                                        output_data[f'{predicted_col}_confidence'][sample_idx] = conf
-                                        break
                                 else:
-                                    output_data[f'{predicted_col}_confidence'][sample_idx] = 0.005
+                                    conf_candidates = list(range(20)) + list(range(20, 100, 10))
+                                    all_ranges = np.array(
+                                        [icps[frozenset(group)].predict(X.values, significance=s / 100)
+                                         for s in conf_candidates])
+                                    all_confs = np.swapaxes(np.swapaxes(all_ranges, 0, 2), 0, 1)
+                                    significances = get_categorical_conf(all_confs, conf_candidates)
+
+                                result.loc[X.index, 'significance'] = significances
+
+                    output_data[f'{predicted_col}_confidence'] = result['significance'].tolist()
+                    confs = [[a, b] for a, b in zip(result['lower'], result['upper'])]
+                    output_data[f'{predicted_col}_confidence_range'] = confs
+
         else:
             for predicted_col in self.lmd['predict_columns']:
                 output_data[f'{predicted_col}_confidence'] = [None] * len(output_data[predicted_col])
